@@ -1,9 +1,9 @@
 """Provide functions to add/find contract checkers."""
-import contextlib
 import functools
 import inspect
 import threading
-from typing import cast, Callable, Any, Iterable, Optional, Tuple, List, Mapping, MutableMapping, Dict, Set
+from typing import Callable, Any, Iterable, Optional, Tuple, List, Mapping, \
+    MutableMapping, Dict
 
 import icontract._represent
 from icontract._globals import CallableT
@@ -52,16 +52,16 @@ def _kwargs_from_call(param_names: List[str], kwdefaults: Dict[str, Any], args:
     :return: resolved arguments as they would be passed to the function
     """
     # pylint: disable=too-many-arguments
-    mapping = dict()  # type: MutableMapping[str, Any]
+    resolved_kwargs = dict()  # type: MutableMapping[str, Any]
 
     # Set the default argument values as condition parameters.
     for param_name, param_value in kwdefaults.items():
-        mapping[param_name] = param_value
+        resolved_kwargs[param_name] = param_value
 
-    # Override the defaults with the values actually suplied to the function.
+    # Override the defaults with the values actually supplied to the function.
     for i, func_arg in enumerate(args):
         if i < len(param_names):
-            mapping[param_names[i]] = func_arg
+            resolved_kwargs[param_names[i]] = func_arg
         else:
             # Silently ignore call arguments that were not specified in the function.
             # This way we let the underlying decorated function raise the exception
@@ -69,9 +69,9 @@ def _kwargs_from_call(param_names: List[str], kwdefaults: Dict[str, Any], args:
             pass
 
     for key, val in kwargs.items():
-        mapping[key] = val
+        resolved_kwargs[key] = val
 
-    return mapping
+    return resolved_kwargs
 
 
 def _not_check(check: Any, contract: Contract) -> bool:
@@ -241,6 +241,9 @@ def _assert_postcondition(contract: Contract, resolved_kwargs: Mapping[str, Any]
             ("The argument(s) of the postcondition have not been set: {}. "
              "Does the original function define them? Did you supply them in the call?").format(missing_args))
 
+        if 'OLD' in missing_args:
+            msg_parts.append(' Did you decorate the function with a snapshot to capture OLD values?')
+
         raise TypeError(''.join(msg_parts))
 
     condition_kwargs = {
@@ -300,13 +303,11 @@ class _Old:
         return "a bunch of OLD values"
 
 
-_THREAD_LOCAL = threading.local()
-
 # This flag is used to avoid recursively checking contracts for the same function or instance while
 # contract checking is already in progress.
 #
-# The value refers to the id() of the function (preconditions and postconditions) or instance (invariants).
-_THREAD_LOCAL.in_progress = cast(Set[int], set())
+# The key refers to the id() of the function (preconditions and postconditions) or instance (invariants).
+_IN_PROGRESS = threading.local()
 
 
 def decorate_with_checker(func: CallableT) -> CallableT:
@@ -333,23 +334,20 @@ def decorate_with_checker(func: CallableT) -> CallableT:
         if param.default != inspect.Parameter.empty:
             kwdefaults[param.name] = param.default
 
-    def unset_checking_in_progress() -> None:
-        """Mark that the checking of the contract is finished."""
-        _THREAD_LOCAL.in_progress.discard(id(func))
+    id_func = str(id(func))
 
     def wrapper(*args, **kwargs):  # type: ignore
         """Wrap func by checking the preconditions and postconditions."""
         # pylint: disable=too-many-branches
 
-        with contextlib.ExitStack() as exit_stack:
-            exit_stack.callback(unset_checking_in_progress)  # pylint: disable=no-member
-
+        # Use try-finally instead of ExitStack for performance.
+        try:
             # If the wrapper is already checking the contracts for the wrapped function, avoid a recursive loop
             # by skipping any subsequent contract checks for the same function.
-            if id(func) in _THREAD_LOCAL.in_progress:
+            if hasattr(_IN_PROGRESS, id_func):
                 return func(*args, **kwargs)
 
-            _THREAD_LOCAL.in_progress.add(id(func))
+            setattr(_IN_PROGRESS, id_func, True)
 
             preconditions = getattr(wrapper, "__preconditions__")  # type: List[List[Contract]]
             snapshots = getattr(wrapper, "__postcondition_snapshots__")  # type: List[Snapshot]
@@ -381,7 +379,7 @@ def decorate_with_checker(func: CallableT) -> CallableT:
                 raise violation_err  # pylint: disable=raising-bad-type
 
             # Capture the snapshots
-            if postconditions:
+            if postconditions and snapshots:
                 old_as_mapping = dict()  # type: MutableMapping[str, Any]
                 for snap in snapshots:
                     # This assert is just a last defense.
@@ -405,6 +403,9 @@ def decorate_with_checker(func: CallableT) -> CallableT:
                     _assert_postcondition(contract=contract, resolved_kwargs=resolved_kwargs)
 
             return result
+        finally:
+            if hasattr(_IN_PROGRESS, id_func):
+                delattr(_IN_PROGRESS, id_func)
 
     # Copy __doc__ and other properties so that doctests can run
     functools.update_wrapper(wrapper=wrapper, wrapped=func)
@@ -461,19 +462,18 @@ def _decorate_with_invariants(func: CallableT, is_init: bool) -> CallableT:
             instance = _find_self(param_names=param_names, args=args, kwargs=kwargs)
             assert instance is not None, "Expected to find `self` in the parameters, but found none."
 
-            _THREAD_LOCAL.in_progress.add(id(instance))
-
-            def remove_from_in_progress() -> None:
-                """Remove the flag which signals that an invariant is already being checked down the call stack."""
-                _THREAD_LOCAL.in_progress.discard(id(instance))
-
-            with contextlib.ExitStack() as exit_stack:
-                exit_stack.callback(remove_from_in_progress)  # pylint: disable=no-member
+            id_instance = str(id(instance))
+            setattr(_IN_PROGRESS, id_instance, True)
 
+            # ExitStack is not used here due to performance.
+            try:
                 for contract in instance.__class__.__invariants__:
                     _assert_invariant(contract=contract, instance=instance)
 
                 return result
+            finally:
+                if hasattr(_IN_PROGRESS, id_instance):
+                    delattr(_IN_PROGRESS, id_instance)
 
     else:
 
@@ -485,21 +485,15 @@ def _decorate_with_invariants(func: CallableT, is_init: bool) -> CallableT:
             instance = _find_self(param_names=param_names, args=args, kwargs=kwargs)
             assert instance is not None, "Expected to find `self` in the parameters, but found none."
 
-            if id(instance) not in _THREAD_LOCAL.in_progress:
-                _THREAD_LOCAL.in_progress.add(id(instance))
+            id_instance = str(id(instance))
+            if not hasattr(_IN_PROGRESS, id_instance):
+                setattr(_IN_PROGRESS, id_instance, True)
             else:
                 # Do not check any invariants to avoid endless recursion.
                 return func(*args, **kwargs)
 
-            def remove_from_in_progress() -> None:
-                """Remove the flag which signals that an invariant is already being checked down the call stack."""
-                _THREAD_LOCAL.in_progress.discard(id(instance))
-
-            with contextlib.ExitStack() as exit_stack:
-                exit_stack.callback(remove_from_in_progress)  # pylint: disable=no-member
-
-                instance = _find_self(param_names=param_names, args=args, kwargs=kwargs)
-
+            # ExitStack is not used here due to performance.
+            try:
                 for contract in instance.__class__.__invariants__:
                     _assert_invariant(contract=contract, instance=instance)
 
@@ -509,6 +503,9 @@ def _decorate_with_invariants(func: CallableT, is_init: bool) -> CallableT:
                     _assert_invariant(contract=contract, instance=instance)
 
                 return result
+            finally:
+                if hasattr(_IN_PROGRESS, id_instance):
+                    delattr(_IN_PROGRESS, id_instance)
 
     functools.update_wrapper(wrapper=wrapper, wrapped=func)
 
